@@ -1,7 +1,7 @@
 /*
  * axfr.h -- generating AXFR responses.
  *
- * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -20,6 +20,6 @@
 #define AXFR_MAX_MESSAGE_LEN MAX_COMPRESSION_OFFSET
 
 query_state_type answer_axfr_ixfr(struct nsd *nsd, struct query *q);
-query_state_type query_axfr(struct nsd *nsd, struct query *query);
+query_state_type query_axfr(struct nsd *nsd, struct query *query, int wstats);
 
 #endif /* _AXFR_H_ */
