@@ -1,7 +1,7 @@
 /*
  * netio.h -- network I/O support.
  *
- * Copyright (c) 2001-2011, NLnet Labs. All rights reserved.
+ * Copyright (c) 2001-2006, NLnet Labs. All rights reserved.
  *
  * See LICENSE for the license.
  *
@@ -16,7 +16,6 @@
  *
  *   NETIO_EVENT_READ: reading will not block.
  *   NETIO_EVENT_WRITE: writing will not block.
- *   NETIO_EVENT_EXCEPT: an exception occurred.
  *   NETIO_EVENT_TIMEOUT: the timeout expired.
  *
  * A file descriptor must be specified if the handler is interested in
@@ -50,8 +49,6 @@
 
 #include "region-allocator.h"
 
-#define NETIO_SLOW_ACCEPT_TIMEOUT 2 /* in seconds */
-
 /*
  * The type of events a handler is interested in.  These can be OR'ed
  * together to specify multiple event types.
@@ -60,9 +57,7 @@ enum netio_event_types {
 	NETIO_EVENT_NONE    = 0,
 	NETIO_EVENT_READ    = 1,
 	NETIO_EVENT_WRITE   = 2,
-	NETIO_EVENT_EXCEPT  = 4,
-	NETIO_EVENT_TIMEOUT = 8,
-	NETIO_EVENT_ACCEPT  = 16
+	NETIO_EVENT_TIMEOUT = 4,
 };
 typedef enum netio_event_types netio_event_types_type;
 
@@ -108,6 +103,9 @@ struct netio_handler
 	 */
 	int fd;
 
+	/** index of the pollfd array for this handler */
+	int pfd;
+
 	/*
 	 * The time when no events should be checked for and the
 	 * handler should be called with the NETIO_EVENT_TIMEOUT
