@@ -43,6 +43,7 @@
 
 #include "config.h"
 #include <stdio.h>
+#include <stdlib.h>
 #ifdef HAVE_SSL
 #include <sys/types.h>
 #include <unistd.h>
@@ -66,6 +67,7 @@
 #include "util.h"
 #include "tsig.h"
 #include "options.h"
+#include "zonec.h"
 
 static void usage(void) ATTR_NORETURN;
 static void ssl_err(const char* s) ATTR_NORETURN;
@@ -74,6 +76,14 @@ static void ssl_path_err(const char* s, const char *path) ATTR_NORETURN;
 /** timeout to wait for connection over stream, in msec */
 #define NSD_CONTROL_CONNECT_TIMEOUT 5000
 
+int zonec_parse_string(region_type* ATTR_UNUSED(region),
+	domain_table_type* ATTR_UNUSED(domains), zone_type* ATTR_UNUSED(zone),
+	char* ATTR_UNUSED(str), domain_type** ATTR_UNUSED(parsed),
+	int* ATTR_UNUSED(num_rrs))
+{
+	return 0;
+}
+
 /** Give nsd-control usage, and exit (1). */
 static void
 usage()
@@ -580,8 +590,11 @@ int main(int argc, char* argv[])
 	if(argc == 0)
 		usage();
 	if(argc >= 1 && strcmp(argv[0], "start")==0) {
-		if(execl(NSD_START_PATH, "nsd", "-c", cfgfile, 
-			(char*)NULL) < 0) {
+		const char *path;
+		if((path = getenv("NSD_PATH")) == NULL) {
+			path = NSD_START_PATH;
+		}
+		if(execl(path, "nsd", "-c", cfgfile, (char*)NULL) < 0) {
 			fprintf(stderr, "could not exec %s: %s\n",
 				NSD_START_PATH, strerror(errno));
 			exit(1);
